#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>

#include "common.h"
#include "macros.h"
#include "xmalloc.h"
#include "mlex.h"
#include "symbol.h"
#include "keyword.h"
#include "as_tree.h"
#include "verify.h"
#include "guess.h"
#include "misc.h"

#define	INIT_ARRAY_SIZE	8

#define note_msg(fmt, arg...) \
do { \
	message(2, mlex_file_name(), mlex_line_number(), fmt, ##arg); \
} while (0)

#define err_msg(fmt, arg...) \
do { \
	errors++; \
	message(1, mlex_file_name(), mlex_line_number(), fmt, ##arg); \
} while (0)

#define warn_msg(fmt, arg...) \
do { \
	message(0, mlex_file_name(), mlex_line_number(), fmt, ##arg); \
} while (0)

#define warn_msg_var_unused(line_number, fmt, arg...) \
do { \
	message(0, mlex_file_name(), line_number, fmt, ##arg); \
} while (0)

#define	MATCH_QUALIFIERS \
( \
	match(TOKEN_MODULE) || match(TOKEN_EXTERN) || \
	match(TOKEN_BYREF) || match(TOKEN_FMT) || \
	match(TOKEN_DECL_STRING) || match(TOKEN_DECL_INT) || \
	match(TOKEN_DECL_REAL) || match(TOKEN_DECL_REAL) || \
	match(TOKEN_DECL_VOID) || match(TOKEN_DECL_CHAR) || \
	match(TOKEN_AMPERSAND) \
)

/* Helper to store id qualifiers. */
struct qualifiers {
	data_type_t	data_type;
	int		is_extern :1;
	int		is_module :1;
	int		is_ref :1;
	int		is_fmt :1;
};
/* A statement helper which travels through nodes. */
struct programme_aux {
	data_type_t	ret_type;
	int		ret_found;
	int		is_cycle;
	int		is_func;
};

extern struct	mlex mlex;

static token_t		current_token;
static struct		mlex mlex_prev;
static unsigned int	errors;		/* count errors, return value */

static struct {
	const char	ch;
	data_type_t	type;
} spec2data[] = {
	{ 'd', DATA_TYPE_INTEGER },
	{ 'x', DATA_TYPE_INTEGER },
	{ 'i', DATA_TYPE_INTEGER },
	{ 'f', DATA_TYPE_REAL },
	{ 's', DATA_TYPE_STRING },
	{ 'c', DATA_TYPE_CHAR },
	{ '0', DATA_TYPE_UNKNOWN }
};

static struct {
	token_t		token;
	data_type_t	type;
} token2data[] = {
	{ TOKEN_DECL_INT,	DATA_TYPE_INTEGER },
	{ TOKEN_DECL_REAL,	DATA_TYPE_REAL },
	{ TOKEN_DECL_CHAR,	DATA_TYPE_CHAR },
	{ TOKEN_DECL_STRING,	DATA_TYPE_STRING },
	{ TOKEN_DECL_VOID,	DATA_TYPE_VOID },
	{ TOKEN_UNKNOWN,	DATA_TYPE_UNKNOWN }
};

static struct {
	token_t		token;
	const char	*oper;
} token2oper[] = {
	{ TOKEN_PLUS,		"+" },
	{ TOKEN_MINUS,		"-" },
	{ TOKEN_ASTERISK,	"*" },
	{ TOKEN_SLASH,		"/" },
	{ TOKEN_PERCENT,	"%" },
	{ TOKEN_LT,		"<" },
	{ TOKEN_GT,		">" },
	{ TOKEN_LE,		"<=" },
	{ TOKEN_GE,		">=" },
	{ TOKEN_NE,		"!=" },
	{ TOKEN_EQ,		"==" },
	{ TOKEN_AND,		"&&" },
	{ TOKEN_OR,		"||" },
	{ TOKEN_UNKNOWN,	NULL }
};

static struct ast_node *stmts();
static struct ast_node *stmt();
static struct ast_node *or_expr();
static struct ast_node *complete_rel_op(struct ast_node *left);

static void check_unused_variables(struct symbol_table *table);
static void parse_qualifiers(struct qualifiers *q);
/* 
	stmts -> stmt stmts_rest 
	stmts_rest -> stmt stmts_rest | NULL
	stmt ->	  scope_expr
		| end_scope_expr
	 	| if_expr
		| for_expr
		| while_expr
		| break_expr
		| continue_expr
		| return_expr
		| declaration_expr
		| other_expr
		| unknown_expr

	or_expr -> and_expr rest_or
	rest_or -> '||' rest_or
	
	and_expr -> rel_expr rest_and
	rest_and -> '&&' rest_and

	expr -> id '=' expr | or_expr

	summ_expr -> mult_expr rest_summ
	rest_summ -> '+' mult_expr rest_summ | '-' mult_expr rest_summ | NULL

	mult_expr -> term_expr rest_mult
	rest_mult -> '*' term_expr rest_mult | '/' term_expr rest_mult | '%' term_expr rest_mult | NULL

	term_expr -> term | '(' expr ')'

	term -> integer | real | id | id_or_indexed

	id_or_indexed -> id | indexed_id

	indexed_id -> id '[' expr ']' rest_indexed_id
	rest_indexed_id -> '[' expr ']' rest_indexed_id | NULL
*/


static void mlex_prev_free()
{
	switch (mlex_prev.token) {
	case TOKEN_ID:
		xfree(mlex_prev.id);
		break;
	case TOKEN_STRING:
		xfree(mlex_prev.str);
		break;
	default:
		break;
	}
}

static void mlex_dup()
{
	switch (mlex.token) {
	case TOKEN_ID:
		mlex_prev.id = xstrdup(mlex.id);
		break;
	case TOKEN_STRING:
		mlex_prev.str = xstrdup(mlex.str);
		break;
	case TOKEN_INTEGER:
		mlex_prev.integer = mlex.integer;
		break;
	case TOKEN_REAL:
		mlex_prev.real = mlex.real;
		break;
	case TOKEN_CHAR:
		mlex_prev.ch = mlex.ch;
		break;
	default:
		break;
	}

	mlex_prev.token = mlex.token;
}

static int match(token_t token)
{
	if (current_token == token) {
		mlex_prev_free();
		mlex_dup();
		current_token = mlex_get_next_token();
		return 1;
	}

	return 0;
}

static void consume_token()
{
	mlex_prev_free();
	mlex_dup();
	current_token = mlex_get_next_token();
}

static void sync_stream()
{
	while ((current_token = mlex_get_next_token()) != TOKEN_EOF) {
		if (current_token == TOKEN_RBRACE)
			break;
	}
}

static data_type_t get_data_type_by_spec(char ch)
{
	unsigned int i;

	for (i = 0; spec2data[i].type != DATA_TYPE_UNKNOWN; i++) {
		if (spec2data[i].ch == ch)
			return spec2data[i].type;
	}

	return DATA_TYPE_UNKNOWN;
}

static data_type_t get_data_type_by_token(token_t token)
{
	unsigned int i;

	for (i = 0; token2data[i].type != DATA_TYPE_UNKNOWN; i++) {
		if (token2data[i].token == token)
			return token2data[i].type;
	}

	return DATA_TYPE_UNKNOWN;
}

static const char *get_oper_type_by_token(token_t token)
{
	unsigned int i;

	for (i = 0; token2oper[i].oper != NULL; i++) {
		if (token2oper[i].token == token)
			return token2oper[i].oper;
	}

	return NULL;
}
/**************/
static int check_fmt_string(struct ast_node_func_call *call,
			    struct symbol_func *symbol)
{
	struct ast_node_const *_const;
	data_type_t *fmtargs;
	unsigned int nfmtargs, i, j;
	boolean_t ok;
	char *ch;

	if (call->args[symbol->nargs-1]->type != NODE_TYPE_CONST)
		return 0;

	_const = (struct ast_node_const *)call->args[symbol->nargs-1];

	fmtargs = NULL;
	nfmtargs = 0;
	ch = _const->str;

	while ((ch = strchr(ch, '%')) != NULL) {
		data_type_t type;
		if (ch[1] == '%') {
			ch += 2;
			continue;
		}
		++ch;
		
		type = get_data_type_by_spec(*ch);
		if (type == DATA_TYPE_UNKNOWN) {
			err_msg("unknown format string specifier %c", *ch);
			goto free;
		}

		nfmtargs++;
		fmtargs = xrealloc(fmtargs, sizeof(data_type_t)*nfmtargs);
		fmtargs[nfmtargs-1] = type;
	}

	if (nfmtargs < (call->nargs - symbol->nargs)) {
		err_msg("too many arguments for format string");
		goto free;
	} else if (nfmtargs > (call->nargs - symbol->nargs)) {
		err_msg("too few arguments for format string");
		goto free;
	}

	for (i = symbol->nargs, j = 0; i < call->nargs; i++, j++) {
		ok = verify_expr_type(call->args[i], fmtargs[j]);
		if (!ok) {
			note_msg("argument %d type not fit format string", i+1);
			goto free;
		}
	}

	if (fmtargs != NULL)
		xfree(fmtargs);

	return 0;
free:
	if (fmtargs != NULL)
		xfree(fmtargs);

	return 1;
}

static int parse_func_call_args(struct ast_node_func_call *call)
{
	struct ast_node *arg_expr;

	while (!match(TOKEN_RPARENTH)) {

		if (match(TOKEN_EOF)) {
			err_msg("reach the end of the file");
			return 1;
		}

		arg_expr = or_expr();
		if (arg_expr == NULL) {
			err_msg("function argument expression");
			return 1;
		}

		ast_node_func_call_add_arg(call, arg_expr);

		if (current_token != TOKEN_RPARENTH && !match(TOKEN_COMMA)) {
			err_msg("expected comma or right parenthesis");
			return 1;
		}
	}

	return 0;
}
/* If arguments are empty, we set default values for them. */
static int set_initval_func_call_args(struct ast_node_func_call *call,
				      struct symbol_func *symbol)
{
	struct symbol_arg *arg;
	struct ast_node *initval;
	unsigned int i;

	for (i = call->nargs; i < symbol->nargs; i++) {
		arg = symbol->args[i];
		if (arg->is_ref)
			err_msg("reference can't be empty, argument %d", i+1);

		initval = arg->initval;
		ast_node_ref(AST_VAR_INIT(initval)->expr);
		ast_node_func_call_add_arg(call, AST_VAR_INIT(initval)->expr);
	}

	return 0;
}

static int check_func_call_args(struct ast_node_func_call *call,
				struct symbol_func *symbol)
{
	struct symbol_arg *arg;
	unsigned int i;
	boolean_t ok;
	int ret;

	/* Check arguments data type. */
	for (i = 0; i < symbol->nargs; i++) {
		arg = symbol->args[i];

		ok = verify_expr_type(call->args[i],
				      SYMBOL(arg)->decl_type);
		if (!ok) {
			note_msg("argument %d data type", i+1);
			return 1;
		}
		/* Check reference node type. */
		if (arg->is_ref) {
			switch (call->args[i]->type) {
			case NODE_TYPE_REFER:
			case NODE_TYPE_ID:
				break;
			default:
				err_msg("not reference type, argument %d", i+1);
				return 1;
			}
		}
	}
	/* In case of the format string function, we can have more call arguments
	 * than the symbol arguments number. */
	if (symbol->with_fmt) {
		/* Check format string arguments. */
		ret = check_fmt_string(call, symbol);
		if (ret != 0)
			return 1;
	} else if (call->nargs > symbol->nargs) {
		err_msg("to many arguments");
		return 1;
	}

	return 0;
}

static struct ast_node *func_call(struct symbol *symbol)
{
	struct ast_node_func_call *call;
	int ret;

	call = ast_node_func_call(symbol->name, symbol->decl_type);

	ret = parse_func_call_args(call);
	if (ret != 0)
		goto free;

	ret = set_initval_func_call_args(call, SYMBOL_FUNC(symbol));
	if (ret != 0)
		goto free;

	ret = check_func_call_args(call, SYMBOL_FUNC(symbol));
	if (ret != 0)
		goto free;

	symbol_incr_reference(symbol);

	return AST_NODE(call);

free:
	sync_stream();
	ast_node_unref(AST_NODE(call));
	return ast_node_stub();
}

static struct ast_node *access_node(struct symbol *symbol)
{
	struct ast_node_access *access;
	struct ast_node *idx_expr;

	access = ast_node_access(symbol->decl_type, symbol->name,
				 SYMBOL_ARRAY(symbol)->dims);
	/* Process indexes, they can't be empty. */
	do {
		if (match(TOKEN_EOF)) {
			err_msg("reach the end of the file");
			goto free;
		}
		/* idx can be any kind of node. */
		idx_expr = or_expr();
		if (idx_expr == NULL) {
			err_msg("index expression");
			goto free;
		}

		ast_node_access_add_idx(access, idx_expr);
		if (!match(TOKEN_RBRACKET)) {
			err_msg("unclosed brackets");
			goto free;
		}
	} while (match(TOKEN_LBRACKET));

	symbol_incr_reference(symbol);

	return AST_NODE(access);

free:
	sync_stream();
	ast_node_unref(AST_NODE(access));

	return ast_node_stub();
}

static struct ast_node *id_node(struct symbol *symbol)
{
	struct ast_node_id *node;

	node = ast_node_id(symbol->decl_type, symbol->name);
	symbol_incr_reference(symbol);

	return AST_NODE(node);
}

static struct ast_node *refer_node(struct symbol *symbol)
{
	struct ast_node_refer *node;

	node = ast_node_refer(symbol->decl_type, symbol->name);
	symbol_incr_reference(symbol);

	return AST_NODE(node);
}

static struct ast_node *token_id(const char *id)
{
	struct ast_node *node;
	struct symbol *symbol;
		
	return_val_if_fail(id, NULL);

	/* Locate symbol in tables. User had to declare it earlier. */
	symbol = symbol_table_lookup_all(id);
	if (symbol == NULL) {
		err_msg("undeclared symbol `%s'", id);
		goto err;
	}
	/* Function call. */
	if (match(TOKEN_LPARENTH)) {
		
		if (symbol->type != SYMBOL_TYPE_FUNC &&
		    symbol->type != SYMBOL_TYPE_FUNC_DECL) {
			err_msg("`%s' was declared as not function", symbol->name);
			goto err;
		}

		node = func_call(symbol);
	} else if (match(TOKEN_LBRACKET)) {
		/* Array indexes. */
		if (symbol->type != SYMBOL_TYPE_ARRAY) {
			err_msg("`%s' was declared as not array", symbol->name);
			goto err;
		}

		node = access_node(symbol);
	} else {
		/* Reference or variable. */
		if (symbol->type != SYMBOL_TYPE_VARIABLE &&
		    symbol->type != SYMBOL_TYPE_ARGUMENT) {
			err_msg("`%s' was declared as not variable", symbol->name);
			goto err;
		}

		if (symbol->type == SYMBOL_TYPE_ARGUMENT && SYMBOL_ARG(symbol)->is_ref) 
			node = refer_node(symbol);
		else
			node = id_node(symbol);
	}

	return node;
err:
	sync_stream();
	return ast_node_stub();
}

#ifdef _MEM_AREA
static struct ast_node *memory_area()
{
	struct ast_node_mem_area *mem_area;
	struct ast_node *idx;
	boolean_t ok;
	char *id, *pos;
	char mem_type, data_type;
	unsigned int offset, nbits;
	int ret, is_unsign, top_watermark;

	if (!match(TOKEN_ID)) {
		err_msg("area name type");
		sync_stream();
		return ast_node_stub();
	}

	idx = NULL;
	id = mlex_prev.id;

	mem_type = id[0];
	if (mem_type != 'I' && mem_type != 'Q') {
		err_msg("unknown memory area type name `%c'", mem_type);
		sync_stream();
		return ast_node_stub();
	}

	++id;
	pos = id;

	while(isdigit(*pos))
		++pos;

	if (pos == id) {
		err_msg("unknown offset in the area", mem_type);
		sync_stream();
		return ast_node_stub();
	}

	ret = str_to_uint(id, &id, &offset, 10);
	if (ret != 0) {
		err_msg("memory area offset");
		sync_stream();
		return ast_node_stub();
	}

	top_watermark = 0;
	is_unsign = 0;

	switch (id[0]) {
	case 'U':
		is_unsign = 1;
		switch (id[1]) {
		case 'B':
			top_watermark = 7;
			break;
		case 'S':
			top_watermark = 15;
			break;
		case 'L':
			top_watermark = 31;
			break;
		default:
			err_msg("memory area data type");
			sync_stream();
			return ast_node_stub();
		}
		data_type = id[1];
		++id;
		break;
	case 'B':
		top_watermark = 7;
		data_type = id[0];
		break;
	case 'S':
		top_watermark = 15;
		data_type = id[0];
		break;
	case 'F':
	case 'L':
		top_watermark = 31;
		data_type = id[0];
		break;
	default:
		err_msg("memory area data type");
		sync_stream();
		return ast_node_stub();
	}

	mem_area = ast_node_mem_area(mem_type, offset, data_type, is_unsign);

	++id;
	if (match(TOKEN_LBRACKET)) {
		idx = or_expr();
		if (idx == NULL) {
			err_msg("index expression");
			sync_stream();
			goto free;
		}

		ok = verify_expr_type(idx, DATA_TYPE_INTEGER);
		if (!ok)
			note_msg("array index data type");

		ast_node_mem_area_add_idx(mem_area, idx);

		if (!match(TOKEN_RBRACKET)) {
			err_msg("unclosed array brackets");
			sync_stream();
			goto free;
		}

		if (match(TOKEN_LBRACKET)) {
			err_msg("memory area is indexed only as one dimension array");
			sync_stream();
			goto free;
		}
	}

	if (match(TOKEN_DOT)) {
		if (!match(TOKEN_ID)) {
			err_msg("bit number");
			sync_stream();
			goto free;
		}

		id = mlex_prev.id;
		if (id[0] != 'B') {
			err_msg("bit number");
			sync_stream();
			goto free;
		}

		id++;

		ret = str_to_uint(id, &id, &nbits, 10);
		if (ret != 0) {
			err_msg("bit number");
			sync_stream();
			goto free;
		}
		
		if (nbits > top_watermark) {
			err_msg("bit number can't exceed %d", top_watermark);
			sync_stream();
			goto free;
		}

		ast_node_mem_area_add_nbits(mem_area, nbits);
	}

	if (*id != 0)
		err_msg("unknown token `%s' memory area", id);

	return AST_NODE(mem_area);

free:
	if (idx != NULL)
		ast_node_unref(idx);

	ast_node_unref(AST_NODE(mem_area));

	return ast_node_stub();
}
#endif

static struct ast_node *term()
{
	struct ast_node *node;

	node = NULL;

	if (match(TOKEN_ID)) {
		node = token_id(mlex_prev.id);
		return node;
	} else if (match(TOKEN_INTEGER)) {
		node = (struct ast_node *)ast_node_const(DATA_TYPE_INTEGER,
							&mlex_prev.integer);
		return node;
	} else if (match(TOKEN_REAL)) {
		node = (struct ast_node *)ast_node_const(DATA_TYPE_REAL,
							&mlex_prev.real);
		return node;
	} else if (match(TOKEN_CHAR)) {
		node = (struct ast_node *)ast_node_const(DATA_TYPE_CHAR,
							&mlex_prev.ch);
		return node;
	} else if (match(TOKEN_STRING)) {
		node = (struct ast_node *)ast_node_const(DATA_TYPE_STRING,
							mlex_prev.str);
		return node;
	}
#ifdef _MEM_AREA
	else if (match(TOKEN_DOLLAR)) {
		node = memory_area();
		return node;
	}
#endif
	return node;
}

static struct ast_node *term_expr()
{	
	struct ast_node *node;
	
	if (match(TOKEN_LPARENTH)) {
		node = or_expr();

		if (!match(TOKEN_RPARENTH)) {
			err_msg("unclosed bracket");
			sync_stream();
		}
	} else {
		node = term();
	}

	return node;
}

/* Now we have just one unary operator '!'. */
static struct ast_node *unary_expr()
{
	struct ast_node *node, *expr;

	if (match(TOKEN_NOT)) {
		expr = term_expr();

		if (expr == NULL) {
			err_msg("not expression is empty");
			sync_stream();
			return ast_node_stub();
		}

		expr = complete_rel_op(expr);
		node = (struct ast_node *)ast_node_not(expr);
	} else {
		node =  term_expr();
	}

	return node;
}

static struct ast_node *rest_mult(struct ast_node *node)
{
	struct ast_node *prev_node, *ret_node, *expr_node;
	const char *oper;

	prev_node = node;

	for (;;) {
		if (!match(TOKEN_ASTERISK) &&
		    !match(TOKEN_SLASH) &&
		    !match(TOKEN_PERCENT))
			break;

		oper = get_oper_type_by_token(mlex_prev.token);

		expr_node = unary_expr();
		if (expr_node == NULL) {
			err_msg("rvalue of `%s'", oper); 
			goto err;
		}
	
		prev_node = (struct ast_node *)ast_node_op(oper, prev_node,
							   expr_node);
        }

        ret_node = prev_node;
     
        return ret_node;

err:
	sync_stream();
	return ast_node_stub();
}

static struct ast_node *mult_expr()
{
	struct ast_node *unary_node;
	struct ast_node *rest_node;
	
	unary_node = unary_expr();
	if (unary_node == NULL)
		return NULL;

	rest_node = rest_mult(unary_node);

	return rest_node;	
}

static struct ast_node *rest_summ(struct ast_node *node)
{
	struct ast_node *prev_node, *ret_node, *expr_node;
	const char *oper;

	prev_node = node;

	for (;;) {
		if (!match(TOKEN_PLUS) &&
		    !match(TOKEN_MINUS))
			break;
 
		oper = get_oper_type_by_token(mlex_prev.token);

		expr_node = mult_expr();
		if (expr_node == NULL) {
			err_msg("rvalue of `%s'", oper);
			goto err;
		}
		prev_node = (struct ast_node *)ast_node_op(oper, prev_node,
							   expr_node);
	}

	ret_node = prev_node;
	
	return ret_node;

err:
	sync_stream();
	return ast_node_stub();
}

static struct ast_node *summ_expr()
{
	struct ast_node *mult_node;
	struct ast_node *rest_node;
	
	mult_node = mult_expr();
	if (mult_node == NULL)
		return NULL;

	rest_node = rest_summ(mult_node);
	
	return rest_node;
}

static struct ast_node *rel_expr()
{
	struct ast_node *left, *right;
	struct ast_node_op *rel_node;
	const char *oper;
	
	left = summ_expr();
	if (left == NULL)
		return NULL;

	if (match(TOKEN_LT) || match(TOKEN_GT) || match(TOKEN_LE) ||
	    match(TOKEN_GE) || match(TOKEN_NE) || match(TOKEN_EQ)) {
		oper = get_oper_type_by_token(mlex_prev.token);

		right = summ_expr();
		if (right == NULL) {
			err_msg("rvalue of `%s'", oper);
			goto err;
		}
		rel_node = ast_node_op(oper, left, right);
	} else {
		return left;
	}

	return AST_NODE(rel_node);

err:
	sync_stream();
	return ast_node_stub();
}

static struct ast_node *complete_rel_op(struct ast_node *left)
{
	struct ast_node_op *rel_node;
	struct ast_node *right;
	data_type_t type;

	int integer = 0;
	float real = 0.0;
	char ch = 0;
	char *str = "";

	switch (left->type) {
	case NODE_TYPE_REL_OP:
	case NODE_TYPE_OR_OP:
	case NODE_TYPE_AND_OP:
	case NODE_TYPE_NOT:
		return left;
	default:
		break;
	}

	type = guess_data_type(left);

	switch (type) {
	case DATA_TYPE_INTEGER:
		right = (struct ast_node *)ast_node_const(type, &integer);
		break;
	case DATA_TYPE_REAL:
		right = (struct ast_node *)ast_node_const(type, &real);
		break;
	case DATA_TYPE_CHAR:
		right = (struct ast_node *)ast_node_const(type, &ch);
		break;
	case DATA_TYPE_STRING:
		right = (struct ast_node *)ast_node_const(type, str);
		break;
	default:
		right = ast_node_stub();
		break;
	}

	rel_node = ast_node_op(">", left, right);

	return AST_NODE(rel_node);
}

static struct ast_node *rest_and(struct ast_node *node)
{
	struct ast_node *expr_node, *ret_node, *prev_node;
	const char *oper;

	prev_node = NULL;

	for (;;) {
		if (!match(TOKEN_AND))
			break;

		if (prev_node == NULL)
			prev_node = complete_rel_op(node);

		oper = get_oper_type_by_token(TOKEN_AND);
		expr_node = rel_expr();

		if (expr_node == NULL) {
			err_msg("rvalue of `%s'", oper); 
			sync_stream();
			expr_node = ast_node_stub();
		}

		expr_node = complete_rel_op(expr_node);
		prev_node = (struct ast_node *)ast_node_op(oper, prev_node,
							   expr_node);
        }

        ret_node = prev_node;

        return ret_node;
}

static struct ast_node *and_expr()
{
	struct ast_node *rel_node;
	struct ast_node *rest_node;

	rel_node = rel_expr();

	if (rel_node == NULL)
		return NULL;

	rest_node = rest_and(rel_node);
	if (rest_node == NULL)
		return rel_node;

	return rest_node;
}

static struct ast_node *rest_or(struct ast_node *node)
{
	struct ast_node *expr_node, *ret_node, *prev_node;
	const char *oper;

	prev_node = NULL;
	for (;;) {
		if (!match(TOKEN_OR))
			break;

		if (prev_node == NULL)
			prev_node = complete_rel_op(node);
			
		oper = get_oper_type_by_token(TOKEN_OR);
		expr_node = and_expr();

		if (expr_node == NULL) {
			err_msg("rvalue of `%s'", oper); 
			sync_stream();
			expr_node = ast_node_stub();
		}
	
		expr_node = complete_rel_op(expr_node);
		prev_node = (struct ast_node *)ast_node_op(oper, prev_node,
							   expr_node);
        }

	ret_node = prev_node;

	return ret_node;
}

static struct ast_node *or_expr()
{
	struct ast_node *and_node;
	struct ast_node *rest_node;

	and_node = and_expr();
	if (and_node == NULL)
		return NULL;

	rest_node = rest_or(and_node);
	if (rest_node == NULL)
		return and_node;

	return rest_node;
}

static struct ast_node *expr()
{
	struct ast_node_assign *assign;
	struct ast_node *left, *right;
	boolean_t ok;
	data_type_t type;

	assign = NULL;	
	right = NULL;
	left = term();

/*	if (left == NULL) {
		err_msg("assignment lvalue");
		goto err;
	} */

	while (match(TOKEN_EQUALITY)) {
		if (assign == NULL)
			assign = ast_node_assign();

		ast_node_assign_add_left(assign, left);
		right = or_expr();
		if (right == NULL) {
			err_msg("assignment rvalue");
			goto err;
		}
		left = right;
	}
	/* It can be a single function call. */
	if (right == NULL)
		return left;

	ast_node_assign_set_right(assign, right);

	type = guess_data_type(AST_NODE(assign));
	if (type == DATA_TYPE_UNKNOWN) {
		err_msg("unknown data type in assignment expression");
		goto err;
	}
	/* Check types in this sub-tree. */
	ok = verify_expr_type(AST_NODE(assign), type);
	if (!ok)
		note_msg("assignment expression");

	return AST_NODE(assign);

err:
	ast_node_unref(AST_NODE(assign));
	sync_stream();
	return ast_node_stub();	
}

static struct ast_node *argument_initializer()
{
	struct ast_node_var_init *arg_init;
	struct ast_node *or_node;

	arg_init = NULL;
	if (match(TOKEN_EQUALITY)) {
		or_node = or_expr();
		if (or_node == NULL) {
			err_msg("argument initialization rvalue");
			return NULL;
		}
		arg_init = ast_node_arg_init(or_node);
	}

	return AST_NODE(arg_init);
}

static struct ast_node *variable_initializer()
{
	struct ast_node_var_init *var_init;
	struct ast_node *or_node;

	var_init = NULL;
	if (match(TOKEN_EQUALITY)) {
		or_node = or_expr();
		if (or_node == NULL) {
			err_msg("variable initialization rvalue");
			return NULL;
		}
		var_init = ast_node_var_init(or_node);
	}

	return AST_NODE(var_init);
}
/* Parse variables declaration, we also handle variables initialization. */
static struct ast_node *process_variable(const char *id, data_type_t data_type,
					 storage_class_t klass)
{
	struct symbol_var *var;
	struct ast_node *initval, **initvals, *node;
	unsigned int ninitvals, i;
	struct symbol *symbol;
	boolean_t ok;
	char *name;

	name = (char *)id;
	initvals = NULL;
	ninitvals = 0;
	/* The cycle is perfomed at least once. */
	do {
		if (match(TOKEN_EOF)) {
			err_msg("reach the end of file in parsing variables");
			goto free;
		}
		/* Lookup in the current scope. */
		symbol = symbol_table_lookup_top(name);
		if (symbol != NULL) {
			err_msg("`%s' was declared earlier", name);
			goto free;
		}

		var = symbol_variable(name, data_type, klass);
		initval = variable_initializer();

		if (initval != NULL) {
			/* Type checking. */
			ok = verify_expr_type(initval, data_type);
			if (!ok)
				note_msg("variable initialization");

			symbol_ref(SYMBOL(var));
			ast_node_init_set_symbol(initval, SYMBOL(var));
			/* We keep all init nodes in the array. Below we link
			 * nodes together and  build a subtree. Then return the
			 * first element to the main tree. */
			initvals = xrealloc(initvals,
				sizeof(struct ast_node *)*(ninitvals+1));
			initvals[ninitvals++] = ast_node_stmt(initval);
		}

		symbol_table_put_symbol(SYMBOL(var));

		if (current_token != TOKEN_SEMICOLON &&
		    current_token != TOKEN_COMMA) {
			err_msg("variable declaration");
			goto free;
		}

		if (match(TOKEN_SEMICOLON))
			break;

		if (match(TOKEN_COMMA) && !match(TOKEN_ID)) {
			err_msg("expected an identifier in variable declaration");
			goto free;
		}

		name = mlex_prev.id;
	} while (1);

	if (initvals != NULL) {
		for (i = 0; i < ninitvals-1; i++) {
			initvals[i]->next = initvals[i+1];
			initvals[i+1]->parent = initvals[i];
		}
		node = initvals[0];
		xfree(initvals);
	} else
		node = ast_node_stub();

	return node;
free:
	sync_stream();

	if (initvals != NULL) {
		for (i = 0; i < ninitvals; i++)
			ast_node_unref(initvals[i]);
		xfree(initvals);
	}
	return ast_node_stub();
}

static struct ast_node *default_arg_init_node(data_type_t type)
{
	struct ast_node_var_init *arg_init;
	struct ast_node_const *_const;

	int integer = 0;
	float real = 0.0;
	char c = '\0';
	char *str = "";

	switch (type) {
	case DATA_TYPE_CHAR:
		_const = ast_node_const(type, &c);
		break;
	case DATA_TYPE_INTEGER:
		_const = ast_node_const(type, &integer);
		break;
	case DATA_TYPE_REAL:
		_const = ast_node_const(type, &real);
		break;
	case DATA_TYPE_STRING:
		_const = ast_node_const(type, str);
		break;
	default:
		err_msg("can't detect data type");
		_const = (struct ast_node_const *)ast_node_stub();
		break;
	}

	arg_init = ast_node_arg_init(AST_NODE(_const));

	return AST_NODE(arg_init);
}

static int check_arg_qualifiers(struct qualifiers *q)
{
	if (q->data_type == DATA_TYPE_UNKNOWN) {
		err_msg("argument data type");
		return 1;
	}

	if (q->is_fmt && q->data_type != DATA_TYPE_STRING) {
		err_msg("`fmt' qualifier can be used only with string type");
		return 1;
	}

	if (q->is_extern) {
		err_msg("function argument can't have `extern' qualifier");
		return 1;
	}

	if (q->is_module) {
		err_msg("function argument can't have `module' qualifier");
		return 1;
	}

	return 0;
}

static struct ast_node *parse_func_body(struct symbol_func *func,
					struct symbol_func *prev_func)
{
	struct ast_node *func_body;
	struct ast_node_scope *scope;
	struct symbol_table *table;
	struct programme_aux aux;
	struct ast_node *initval;
	data_type_t ret_type;
	unsigned int i;
	boolean_t ok;

	/* If init_node is NULL, need to create a default value. */
	for (i = 0; i < func->nargs; i++) {
		struct symbol_arg *arg = func->args[i];
		if (SYMBOL(arg)->name == NULL) {
			err_msg("argument name can't be empty `%s'", SYMBOL(func)->name);
			goto free;
		}

		if (arg->initval == NULL) {
			initval = default_arg_init_node(SYMBOL(arg)->decl_type);
			symbol_arg_set_initval(arg, initval);
		} else {
			ok = verify_expr_type(arg->initval, SYMBOL(arg)->decl_type);
			if (!ok)
				note_msg("function argument initialization");
		}
	}

	if (prev_func != NULL) {
		symbol_table_remove_symbol(SYMBOL(prev_func)->name);
		symbol_unref(SYMBOL(prev_func));
	}

	symbol_table_put_symbol(SYMBOL(func));

	ret_type = SYMBOL(func)->decl_type;
	memset(&aux, 0, sizeof(struct programme_aux));
	++aux.is_func;
	aux.ret_type = ret_type;
	/* Push a new symbol scope. */
	symbol_table_push();
	/* Add arguments to the new scope */
	for (i = 0; i < func->nargs; i++) {
		symbol_ref(SYMBOL(func->args[i]));
		symbol_table_put_symbol(SYMBOL(func->args[i]));
	}
	/* Parse the function body. */
	func_body = stmts(&aux);
	/* Keep the table of this scope in the node. */
	table = symbol_table_get_current_table();
	check_unused_variables(table);
	scope = ast_node_scope(table, func_body);
	/* Pop a symbol scope */
	symbol_table_pop();
	--aux.is_func;

	if (ret_type != DATA_TYPE_VOID && aux.ret_found == 0)
		err_msg("function must return a value");

	return (struct ast_node *)ast_node_function(SYMBOL(func)->name,
					SYMBOL(func)->decl_type, AST_NODE(scope));
free:
	sync_stream();
	symbol_unref(SYMBOL(func));
	return ast_node_stub();
}

static struct ast_node *parse_func_decl(struct symbol_func *func,
					struct symbol_func *prev_func)
{
	struct symbol_arg *arg;
	unsigned int i;

	/* The last check for declaration. */
	for (i = 0; i < func->nargs; i++) {
		arg = func->args[i];
		if (arg->initval != NULL) {
			err_msg("can't use argument initialization in declaration");
			sync_stream();
			return ast_node_stub();
		}
	}

	if (prev_func != NULL) {
		symbol_table_remove_symbol(SYMBOL(prev_func)->name);
		symbol_unref(SYMBOL(prev_func));
	}

	symbol_table_put_symbol(SYMBOL(func));
	symbol_func_mark_as_decl(func);

	return ast_node_stub();
}

static int parse_func_args(struct symbol_func *func)
{
	struct qualifiers qualifiers;
	struct symbol_arg *arg;
	struct ast_node *initval;
	int ret;

	while (!match(TOKEN_RPARENTH)) {
		/* Can't reach the end of the file. */
		if (match(TOKEN_EOF)) {
			err_msg("reach the end of file parsing arguments");
			goto err;
		}
		/* At last iteration we already found format string
		 * argument and it must be the last one. */
		if (func->with_fmt) {
			err_msg("format string must be the last argument");
			goto err;
		}
		/* Expect arguments qualifiers. */
		if (!MATCH_QUALIFIERS) {
			err_msg("argument qualifiers");
			goto err;
		}
		/* Parse and check argument qualifiers. */
		memset(&qualifiers, 0, sizeof(struct qualifiers));
		parse_qualifiers(&qualifiers);
		ret =  check_arg_qualifiers(&qualifiers);
		if (ret != 0) 
			goto err;
		/* Create argument and fill fields. */
		arg = symbol_argument();
		symbol_arg_set_type(arg, qualifiers.data_type);
		arg->is_fmt = func->with_fmt = qualifiers.is_fmt;
		arg->is_ref = qualifiers.is_ref;
		/* In case of function prototype -- name field can be empty. */
		if (match(TOKEN_ID))
			symbol_arg_set_name(arg, mlex_prev.id);
		/* If found a name and it is no a reference, parse a default
		 * argument value. */
		if (SYMBOL(arg)->name != NULL && !arg->is_ref) {
			initval = argument_initializer(SYMBOL(arg)->name);
			symbol_arg_set_initval(arg, initval);
		}
		/* Time to add the argument to our function symbol. */
		symbol_func_add_arg(func, arg);
		if (current_token != TOKEN_RPARENTH && !match(TOKEN_COMMA)) {
			err_msg("expected comma between arguments");
			goto err;
		}
	}

	return 0;

err:
	sync_stream();
	return 1;
}

static struct ast_node *process_function(const char *id, data_type_t data_type, storage_class_t klass)
{
	struct symbol *symbol;
	struct symbol_func *func, *prev_func;
	struct ast_node *node;
	unsigned int i;
	int ret;

	symbol = symbol_table_lookup_top(id);

	if (symbol != NULL && (symbol->type != SYMBOL_TYPE_FUNC_DECL &&
			       symbol->type != SYMBOL_TYPE_FUNC)) {
		err_msg("`%s' was declared as not function", id);
		sync_stream();
		return ast_node_stub();
	}

	if (symbol != NULL)
		prev_func = SYMBOL_FUNC(symbol);
	else
		prev_func = NULL;

	func = symbol_func(id, data_type, klass);
	/* Get arguments. */
	ret = parse_func_args(func);
	if (ret != 0)
		goto free;
	/* Compare two function symbols, the must be equal. */
	if (prev_func != NULL) {
		if (SYMBOL(prev_func)->klass != SYMBOL(func)->klass) {
			err_msg("conflicting argument qualifiers for `%s'", id);
			goto free;
		}
		if (SYMBOL(prev_func)->decl_type != SYMBOL(func)->decl_type) {
			err_msg("conflicting return data type for `%s'", id);
			goto free;
		}
		if (prev_func->nargs != func->nargs) {
			err_msg("conflicting number of arguments for `%s'", id);
			goto free;
		}
		for (i = 0; i < func->nargs; i++) {
			struct symbol_arg *arg1 = prev_func->args[i];
			struct symbol_arg *arg2 = func->args[i];
			if (SYMBOL(arg1)->decl_type == SYMBOL(arg2)->decl_type)
				continue;
			if (arg1->is_fmt == arg2->is_fmt &&
			    arg1->is_ref == arg2->is_ref)
				continue;
			err_msg("conflicting type of arguments for `%s'", id);
			goto free;
		}
	}
	/* Expect function declaration or function itself. */
	if (match(TOKEN_LBRACE)) {
		if (klass == SYMBOL_VISIB_EXTERN) {
			err_msg("use `extern' with function");
			goto free;
		}
		/* We found function redefinition. */
		if (prev_func != NULL &&
			SYMBOL(prev_func)->type == SYMBOL_TYPE_FUNC) {
			err_msg("redefinition of function `%s'", id);
			goto free;
		}
		node = parse_func_body(func, prev_func);
	} else if (match(TOKEN_SEMICOLON)) {

		node = parse_func_decl(func, prev_func);

	} else {
		err_msg("a left brace or a semicolon is missed");
		goto free;
	}

	return node;

free:
	symbol_unref(SYMBOL(func));
	sync_stream();

	return ast_node_stub();
}

static struct ast_node *array_initializer(unsigned int *dims, const unsigned int ndims)
{
	struct ast_node_array_init *array_init;
	unsigned int *help_dims;
	unsigned int nesting, num, i;

	nesting = num = 0;
	/* create alternative array, we'll compare it with original */
	help_dims = xmalloc0(sizeof(unsigned int)*ndims);
	/* { */
	consume_token();

	array_init = ast_node_array_init();

	for (;;) {
		struct ast_node *initval;

		if (match(TOKEN_EOF)) {
			err_msg("reach the end of file in parsing array initialization");
			sync_stream();
			goto free;
		}

		if (match(TOKEN_LBRACE)) {
			if (match(TOKEN_COMMA)) {
				err_msg("expected expression before `,' token");
				sync_stream();
				goto free;
			}

			help_dims[nesting]++;

			if (dims[nesting] != 0 && help_dims[nesting] > dims[nesting]) {
				err_msg("exceed array declaration size");
				sync_stream();
				goto free;
			}

			nesting++;

			if (nesting == ndims) {
				err_msg("exceed array declaration dimension size");
				sync_stream();
				goto free;
			}

			continue;
		}

		if (match(TOKEN_RBRACE)) {
			/* when brace is closed, we need keep track of absent values
			 * using original and alternative arrays */
			unsigned int dif, mult;

			if (nesting == 0)
				break;

			switch (current_token) {
			case TOKEN_COMMA:
			case TOKEN_RBRACE:
				break;
			default:
				err_msg("must be a comma or a right brace");
				sync_stream();
				goto free;
			}

			match(TOKEN_COMMA);

			if (dims[nesting] != 0) {
				mult = 1;
				dif = dims[nesting] - help_dims[nesting];
				for (i = nesting + 1; i < ndims; i++)
					mult *= dims[i];
				num += dif * mult;
			}

			help_dims[nesting] = 0;
			nesting--;
			continue;
		}
		/* Gathering initialization values one by one */
		help_dims[nesting]++;

		if (dims[nesting] != 0 && help_dims[nesting] > dims[nesting]) {
			err_msg("exceed array declaration size");
			sync_stream();
			goto free;
		}

		initval = or_expr();

		if (initval == NULL) {
			err_msg("unknown expression in array initialization");
			sync_stream();
			goto free;
		}

		ast_node_array_add_initval(array_init, initval, num);
		num++;

		switch (current_token) {
		case TOKEN_COMMA:
		case TOKEN_RBRACE:
			break;
		default:
			err_msg("must be a comma or a right brace");
			sync_stream();
			goto free;
		}

		match(TOKEN_COMMA); /* no two commas */
		if (match(TOKEN_COMMA)) {
			err_msg("expected expression before `,' token");
			sync_stream();
			goto free;
		}
	}

	if (dims[0] != 0 && dims[0] < help_dims[0]) {
		warn_msg("excess elements in array initializer");
		dims[0] = help_dims[0];
	}
	/* The first cell size can not be defined, we calculate it. */
	if (dims[0] == 0)
		dims[0] = help_dims[0];

	xfree(help_dims);

	return AST_NODE(array_init);

free:
	if (help_dims != NULL)
		xfree(help_dims);

	ast_node_unref(AST_NODE(array_init));

	return ast_node_stub();
}
/* Parse an array declaration, we also wait its initialization. Initialization
 * nodes are placed similar to common variables initialization nodes. */
static struct ast_node *process_array(const char *id, data_type_t data_type, storage_class_t klass)
{
	struct symbol_array *array;
	struct ast_node *initval, *node;
	struct symbol *symbol;
	boolean_t ok;
	unsigned int size, i;

	symbol = symbol_table_lookup_top(id);
	if (symbol != NULL) {
		err_msg("identifier `%s' was declared before", id);
		sync_stream();
		return ast_node_stub();
	}

	array = symbol_array(id, data_type, klass);
	symbol_table_put_symbol(SYMBOL(array));
	/* Get an array dimension. */
	initval = NULL;

	do {
		if (match(TOKEN_EOF)) {
			err_msg("reach the end of file parsing an array declaration `%s'", id);
			sync_stream();
			return ast_node_stub();
		}

		switch (current_token) {
		case TOKEN_INTEGER:
		case TOKEN_RBRACKET:
			break;
		default:
			err_msg("wrong index in declaration `%s'", id);
			sync_stream();
			return ast_node_stub();
		}

		if (match(TOKEN_INTEGER))
			size = mlex_prev.integer;
		else
			size = 0;

		symbol_array_add_dim(array, size);

		if (!match(TOKEN_RBRACKET)) {
			err_msg("unclosed array brackets in declaration");
			sync_stream();
			return ast_node_stub();
		}
	} while (match(TOKEN_LBRACKET));

	switch (current_token) {
	case TOKEN_EQUALITY:
	case TOKEN_SEMICOLON:
		break;
	default:
		err_msg("expected semicolon or array initialization");
		sync_stream();
		return ast_node_stub();
	}

	if (match(TOKEN_SEMICOLON)) {
		/* Every declaration indexes must be defined */
		for (i = 0; i < array->ndims; i++) {

			if (array->dims[i] != 0)
				continue;

			err_msg("array index has incomplete type or zero value");
			sync_stream();
		}

		return ast_node_stub();
	}
	/* = */
	consume_token();
	if (current_token != TOKEN_LBRACE) {
		err_msg("expected left brace after `='");
		sync_stream();
		return ast_node_stub();
	}
	/* Like in C the most left index can be empty, we'll calculate it. */
	for (i = 1; i < array->ndims; i++) {

		if (array->dims[i] != 0 )
			continue;

		err_msg("array index has incomplete element type or zero value");
		sync_stream();
		return ast_node_stub();
	}

	initval = array_initializer(array->dims, array->ndims);
	symbol_ref(SYMBOL(array));
	ast_node_init_set_symbol(initval, SYMBOL(array));

	ok = verify_expr_type(initval, data_type);
	if (!ok)
		note_msg("array initialization");

	node = initval;

	if (!match(TOKEN_SEMICOLON)) {
		err_msg("expected semicolon after `}' in array declaration");
		sync_stream();
	}

	return node;
}

static struct ast_node *for_expr(struct programme_aux *aux)
{
	struct ast_node_for *for_node;
	struct ast_node *stmt_node;
	struct ast_node *optexpr1, *optexpr2, *optexpr3;
	data_type_t type;
	boolean_t ok;

	if (!aux->is_func) {
		err_msg("can't use `for' outside function");
		sync_stream();
		return ast_node_stub();
	}

	++aux->is_cycle;

	if (!match(TOKEN_LPARENTH)) {
		err_msg("( expected after `for'");
		sync_stream();
		return ast_node_stub();
	}

	optexpr1 = expr();

	if (!match(TOKEN_SEMICOLON)) {
		err_msg("semicolon expected in `for('");
		sync_stream();
		return ast_node_stub();
	}

	if (optexpr1 != NULL) {
		type = guess_data_type(optexpr1);
		ok = verify_expr_type(optexpr1, type);
		if (!ok)
			note_msg("`for' the first expression data type");
	}

	optexpr2 = or_expr();

	if (!match(TOKEN_SEMICOLON)) {
		err_msg("semicolon expected in `for(;'");
		sync_stream();
		return ast_node_stub();
	}

	if (optexpr2 != NULL) {
		optexpr2 = complete_rel_op(optexpr2);

		type = guess_data_type(optexpr2);
		ok = verify_expr_type(optexpr2, type);
		if (!ok)
			note_msg("`for' condition expression data type");
	}

	optexpr3 = expr();

	if (!match(TOKEN_RPARENTH)) {
		err_msg(") expected after `for (optexpr1; optexp2; optexpr3'");
		sync_stream();
		return ast_node_stub();
	}

	if (optexpr3 != NULL) {
		type = guess_data_type(optexpr3);
		ok = verify_expr_type(optexpr3, type);
		if (!ok)
			note_msg("`for' the third expression data type");
	}

	stmt_node = stmt(aux);

	if (stmt_node == NULL) {
		err_msg("`for' statement body can't be empty");
		sync_stream();
		return ast_node_stub();
	}

	--aux->is_cycle;

	for_node = ast_node_for(optexpr1, optexpr2, optexpr3, stmt_node);

	return AST_NODE(for_node);
}

static struct ast_node *if_expr(struct programme_aux *aux)
{
	struct ast_node_if *if_node;
	struct ast_node *expr_node;
	struct ast_node *stmt_node;
	struct ast_node *else_node;
	data_type_t type;
	boolean_t ok;

	if (!aux->is_func) {
		err_msg("can't use `if' outside function");
		sync_stream();
		return ast_node_stub();
	}

	if (!match(TOKEN_LPARENTH)) {
		err_msg("( expected after `if'");
		sync_stream();
		return ast_node_stub();
	}

	expr_node = or_expr();

	if (expr_node == NULL) {
		err_msg("`if' condition is empty");
		sync_stream();
		return ast_node_stub();
	}

	expr_node = complete_rel_op(expr_node);

	type = guess_data_type(expr_node);

	ok = verify_expr_type(expr_node, type);
	if (!ok)
		note_msg("`if' condition expression data types");

	if (!match(TOKEN_RPARENTH)) {
		err_msg(") expected after 'if (expr'");
		sync_stream();
		return ast_node_stub();
	}

	stmt_node = stmt(aux);

	if (stmt_node == NULL) {
		err_msg("`if' body statement can't be empty ");
		sync_stream();
		return ast_node_stub();
	}

	else_node = NULL;

	if (match(TOKEN_ELSE))
		else_node = stmt(aux);

	if_node = ast_node_if(expr_node, stmt_node, else_node);

	return AST_NODE(if_node);
}

static struct ast_node *while_expr(struct programme_aux *aux)
{
	struct ast_node_while *while_node;
	struct ast_node *expr_node;
	struct ast_node *stmt_node;
	data_type_t type;
	boolean_t ok;

	if (!aux->is_func) {
		err_msg("can't use `while' outside function");
		sync_stream();
		return ast_node_stub();
	}

	++aux->is_cycle;

	if (!match(TOKEN_LPARENTH)) {
		err_msg("( expected after `while'");
		sync_stream();
		return ast_node_stub();
	}

	expr_node = rel_expr();

	if (expr_node == NULL) {
		err_msg("`while' ( expr ) can't be empty");
		sync_stream();
		return ast_node_stub();
	}

	expr_node = complete_rel_op(expr_node);
	type = guess_data_type(expr_node);

	ok = verify_expr_type(expr_node, type);
	if (!ok)
		note_msg("`while' condition expression data types");

	if (!match(TOKEN_RPARENTH)) {
		err_msg(") expected after `while (expr'");
		sync_stream();
		return ast_node_stub();
	}

	stmt_node = stmt(aux);

	if (stmt_node == NULL) {
		err_msg("`while' body statement can't be empty");
		sync_stream();
		return ast_node_stub();
	}

	while_node = ast_node_while(expr_node, stmt_node);

	--aux->is_cycle;

	return AST_NODE(while_node);
}

static struct ast_node *return_expr(struct programme_aux *aux)
{
	struct ast_node_return *return_node;
	struct ast_node *expr_node;
	data_type_t ret_type;
	boolean_t ok;

	if (!aux->is_func) {
		err_msg("can't use `return' outside function");
		sync_stream();
		return ast_node_stub();
	}

	ret_type = aux->ret_type;
	++aux->ret_found;

	expr_node = NULL;
	if (ret_type == DATA_TYPE_VOID) {
		if (!match(TOKEN_SEMICOLON)) {
			err_msg("return void value, expected semicolon");
			sync_stream();
			return ast_node_stub();
		}
		return_node = ast_node_return(expr_node, ret_type);
		return AST_NODE(return_node);
	}

	if (match(TOKEN_SEMICOLON)) {
		err_msg("expected return value");
		sync_stream();
		return ast_node_stub();
	}

	expr_node = or_expr();
	if (expr_node == NULL) {
		err_msg("`return' expression");
		sync_stream();
		return ast_node_stub();
	}

	return_node = ast_node_return(expr_node, ret_type);
	ok = verify_expr_type(AST_NODE(return_node), ret_type);
	if (!ok)
		note_msg("return value data type");

	if (!match(TOKEN_SEMICOLON)) {
		err_msg("expected semicolon after return expression");
		sync_stream();
		return ast_node_stub();
	}

	return AST_NODE(return_node);
}

static struct ast_node *break_expr(struct programme_aux *aux)
{
	struct ast_node_break *break_node;

	if (!aux->is_func) {
		err_msg("can't use `break' outside function");
		sync_stream();
		return ast_node_stub();
	}

	if (!aux->is_cycle) {
		err_msg("can't use `break' outside `while' or `for'");
		sync_stream();
		return ast_node_stub();
	}

	if (!match(TOKEN_SEMICOLON)) {
		err_msg("semicolon expected after `break'");
		sync_stream();
		return ast_node_stub();
	}

	break_node = ast_node_break();

	return AST_NODE(break_node);
}

static struct ast_node *continue_expr(struct programme_aux *aux)
{
	struct ast_node_continue *continue_node;

	if (!aux->is_func) {
		err_msg("can't use `continue' outside function");
		sync_stream();
		return ast_node_stub();
	}

	if (!aux->is_cycle) {
		err_msg("can't use continue outside `while' or `for'");
		sync_stream();
		return ast_node_stub();
	}

	if (!match(TOKEN_SEMICOLON)) {
		err_msg("semicolon expected after `continue'");
		sync_stream();
		return ast_node_stub();
	}

	continue_node = ast_node_continue();

	return AST_NODE(continue_node);
}

static void check_unused_variables(struct symbol_table *table)
{
	struct symbol **symbols;
	unsigned int size, i;

	return_if_fail(table != NULL);

	symbols = symbol_table_extract_variables(table, &size);

	for (i = 0; i < size; i++) {
		switch (symbols[i]->type) {
		case SYMBOL_TYPE_VARIABLE:
		case SYMBOL_TYPE_ARRAY:
			break;
		default:
			continue;
		}

		if (symbols[i]->nr_references == 0) {
			warn_msg_var_unused(symbols[i]->line_number,
				"unused variable `%s'", symbols[i]->name);
		}
	}

	if (symbols != NULL)
		xfree(symbols);
}
/* Find a new namespace, create a new symbol table, parse a new block, pop
 * the namespace's symbol table and create a node. */
static struct ast_node *scope_expr(struct programme_aux *aux)
{
	struct ast_node_scope *scope_node;
	struct ast_node *stmt_node;
	struct symbol_table *table;

	if (!aux->is_func) {
		err_msg("unexpected left brace");
		sync_stream();
		return ast_node_stub();
	}
	/* Push a new symbol scope. */
	symbol_table_push();
	/* Parse a scope. */
	stmt_node = stmts(aux);
	/* Save the table in the node. */
	table = symbol_table_get_current_table();
	check_unused_variables(table);
	scope_node = ast_node_scope(table, stmt_node);
	/* Pop a symbol scope */
	symbol_table_pop();

	return AST_NODE(scope_node);
}

static struct ast_node *declaration_expr(struct qualifiers *q,
					 struct programme_aux *aux)
{
	struct ast_node *node;
	storage_class_t klass;
	char *name;

	name = NULL;

	if (q->is_module && aux->is_func) {
		warn_msg("use `module' in the function");
		q->is_module = 0;
	}

	if (q->is_extern && q->is_module)
		err_msg("`module' and `extern' are mutual exclusions");

	if (q->is_ref)
		err_msg("use `byref' only in function arguments");
	if (q->is_fmt)
		err_msg("use `fmt' only in function arguments");

	if (q->is_extern)
		klass = SYMBOL_VISIB_EXTERN;
	else if (q->is_module)
		klass = SYMBOL_VISIB_MODULE;
	else
		klass = SYMBOL_VISIB_GLOBAL;

	if (q->data_type == DATA_TYPE_UNKNOWN) {
		warn_msg("return type defaults to `int'");
		q->data_type = DATA_TYPE_INTEGER;
	}

	if (!match(TOKEN_ID)) {
		err_msg("expected identifier");
		goto err;
	}

	name = xstrdup(mlex_prev.id);
	/* A function, an array or a variable declaration. */
	if (match(TOKEN_LPARENTH)) {

		node = process_function(name, q->data_type, klass);

	} else if (match(TOKEN_LBRACKET)) {
		if (q->data_type == DATA_TYPE_VOID) {
			err_msg("array `%s' was  declared void", name);
			goto err;
		}
		node = process_array(name, q->data_type, klass);
	} else {
		if (q->data_type == DATA_TYPE_VOID) {
			err_msg("variable `%s' was declared void", name);
			goto err;
		}
		node = process_variable(name, q->data_type, klass);
	}

	xfree(name);

	return node;
err:
	if (name != NULL)
		xfree(name);

	sync_stream();
	return ast_node_stub();
}

static void parse_qualifiers(struct qualifiers *q)
{
	int leave;

	leave = 0;
	q->data_type = DATA_TYPE_UNKNOWN;

	do {
		switch (mlex_prev.token) {
		case TOKEN_MODULE:
			if (q->is_module)
				warn_msg("duplicate `module'");
			q->is_module = 1;
			break;
		case TOKEN_EXTERN:
			if (q->is_extern)
				warn_msg("duplicate `extern'");
			q->is_extern = 1;
			break;
		case TOKEN_BYREF:
		case TOKEN_AMPERSAND:
			if (q->is_ref)
				warn_msg("duplicate reference qualifier");
			q->is_ref = 1;
			break;
		case TOKEN_FMT:
			if (q->is_fmt)
				warn_msg("duplicate `fmt'");
			q->is_fmt = 1;
			break;
		case TOKEN_DECL_INT:
		case TOKEN_DECL_CHAR:
		case TOKEN_DECL_STRING:
		case TOKEN_DECL_REAL:
		case TOKEN_DECL_VOID:
			if (q->data_type != DATA_TYPE_UNKNOWN)
				err_msg("two or more data types in declaration "
					"specifiers");
			q->data_type = get_data_type_by_token(mlex_prev.token);
			break;	
		default:
			leave = 1;
			break;
		}
	} while (!leave && MATCH_QUALIFIERS);
}

static struct ast_node *qualifiers_expr(struct programme_aux *aux)
{
	struct qualifiers q;
	struct ast_node *expr_node;

	memset(&q, 0, sizeof(q));
	parse_qualifiers(&q);

	expr_node = declaration_expr(&q, aux);

	return expr_node;
}

static struct ast_node *other_expr(struct programme_aux *aux)
{
	struct ast_node *expr_node;

	expr_node = expr();

	if (expr_node->type != NODE_TYPE_ASSIGN &&
	    expr_node->type != NODE_TYPE_CALL)
		err_msg("expression do nothing");

	if (!match(TOKEN_SEMICOLON))
		err_msg("expected semicolon");

	return expr_node;
}

static struct ast_node *unknown_expr(struct programme_aux *aux)
{
	struct ast_node *node;

	err_msg("unknown token");
	sync_stream();

	node = ast_node_stub();

	return node;
}

static struct ast_node *stmt(struct programme_aux *aux)
{
	struct ast_node *expr_node;

	if (match(TOKEN_LBRACE)) {

		expr_node = scope_expr(aux);

		return ast_node_stmt(expr_node);
	} else if (match(TOKEN_RBRACE)) {
		/* TODO: check for right brace out ot function. */
		expr_node = ast_node_scope_end();

		return expr_node;
	} else if (match(TOKEN_IF)) {

		expr_node = if_expr(aux);

		return ast_node_stmt(expr_node);
	} else if (match(TOKEN_FOR)) {
		
		expr_node = for_expr(aux);
		
		return ast_node_stmt(expr_node);
	} else if (match(TOKEN_WHILE)) {
		
		expr_node = while_expr(aux);

		return ast_node_stmt(expr_node);
	} else if (MATCH_QUALIFIERS) {

		expr_node = qualifiers_expr(aux);

		return ast_node_stmt(expr_node);
	} else if (match(TOKEN_RETURN)) {

		expr_node = return_expr(aux);

		return ast_node_stmt(expr_node);
	} else if (match(TOKEN_CONTINUE)) {

		expr_node = continue_expr(aux);

		return ast_node_stmt(expr_node);
	} else if(match(TOKEN_BREAK)) {

		expr_node = break_expr(aux);

		return ast_node_stmt(expr_node);
	} else if (match(TOKEN_SEMICOLON)) {

		expr_node = ast_node_stub();

		return ast_node_stmt(expr_node);
	} else if (match(TOKEN_UNKNOWN)) {

		expr_node = unknown_expr(aux);

		return ast_node_stmt(expr_node);
	} else {

		expr_node = other_expr(aux);
			
		return ast_node_stmt(expr_node);
	}

	return NULL;
}

static struct ast_node *stmts(struct programme_aux *aux)
{
	struct ast_node *stmt_node;
	struct ast_node *iter_node;
	struct ast_node *stmts_head;
	
	iter_node = NULL;
	stmts_head = NULL;

	while (current_token != TOKEN_EOF) {

		stmt_node = stmt(aux);

		if (stmts_head == NULL)
			stmts_head = stmt_node;
				
		if (iter_node != NULL) {
			iter_node->next = stmt_node;
			stmt_node->parent = iter_node;
		}

		iter_node = stmt_node;

		if (stmt_node->type == NODE_TYPE_SCOPE_END) {
			break;
		}
	}

	return stmts_head;
}

unsigned int programme(struct ast_node_root **root)
{
	struct ast_node *tree;
	struct ast_node_scope *scope;
	struct symbol_table *global;
	struct programme_aux aux;
	
	consume_token();
	memset(&aux, 0, sizeof(struct programme_aux));

	tree = stmts(&aux);
	mlex_prev_free();

	global = symbol_table_get_global_table();
	check_unused_variables(global);

	scope = ast_node_scope(global, tree);
	*root = ast_node_root(AST_NODE(scope));

	return errors;
}

